#!/usr/bin/env python
"""
 Rapid prototype of metadata, data and control flow for the new radiometric calibration module
 This calibration model is applicable to SHIS FBF directories as is.

 Currently, a linear interpolation synthetic blackbody spectrum model is being used 
 (for each of hot+fwd, ambient+fwd hot+bwd, ambient+bwd), and the blackbody temperatures are simply
 taken from the time of observation. Both of these simplifying assumptions can be easily replaced with
 more sophisticated mathematical models.

 author: Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 version: $Id: CalProt.py,v 1.15 2005/03/04 23:05:56 maciek Exp $

 Copyright 2004 University of Wisconsin Madison, Space Science and Engineering Center

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from Numeric import *   # Python Numeric array library (todo: replace with NumArray)
from FBF import *       # SSEC's homebrew flat binary format

DEBUG_ON = 0
GPLOT_ON = 1

SPEC_TYPES = ('ABB', 'HBB', 'OBS', 'SKY', '?')  # Ambient BlackBody, Hot BlackBody, Observation, Sky
FTS_DIRECTIONS = ('FW', 'BW')                   # Forward, Backward (piston scan direction)
CAL_KEYS = ('FWABB', 'BWABB', 'FWHBB', 'BWHBB') # keys for the four kinds of blackbody contexts


# These planck definitions should be elsewhere. But I want this demo as self-contained as possible
h = 6.62606876E-34  # Planck constant in Js
c = 2.99792458E8   # photon speed in m/s
k = 1.3806503E-23   # Boltzmann constant in J/K

c1 = 2*h*c*c*1e8
c2 = h*c/k*1e2

class Planck:
    # initialize with a wavenumber scale.
    def __init__(self, wn):
        self.wn = wn
        
    # takes radiance spectrum, returns brightness-temperature spectrum
    def rad2bt(self, rad):
        return c2 * freq / ( log(1 + c1 * (self.wn ** 3) / (rad / 1e3)))
    
    # inverse of rad2bt
    def bt2rad(self, bt):
        return 1e3 * c1 * (self.wn ** 3) /(exp((c2 * self.wn) / bt) - 1)
    
    # gives a perfect blackbody curve for a given temperature
    def perfectBb(self, temp):
        return self.bt2rad(ones(len(self.wn)) * temp)


# context superclass. defines begin/end times for applicability of model
class Context:
    def __init__(self, beginT, endT):
        self.beginT = beginT
        self.endT = endT
    def isValid(self, time):
        return time >= self.beginT and time <= self.endT


# linear blackbody context
# basically, time end-points (beginT, endT) and two spectra (beginS, endS)
class LinBbContext(Context):
    def __init__(self, beginT, endT, beginS, endS, cxtType):
        Context.__init__(self, beginT, endT)
        self.cxtType = cxtType
        self.beginS = beginS
        self.endS = endS
        self.slope = (endS - beginS)/(endT - beginT)
        
    def getSpectrumAt(self,time):
        if DEBUG_ON:
            ##### DEBUG #####
            if not self.isValid(time):
                print self.cxtType,
                if time < self.beginT:
                    print "%s <", self.beginT - time,
                elif time > self.endT:
                    print "> %s", time - self.endT,
            ##### GUBED #####
            
        return self.beginS + self.slope * (time - self.beginT)            
            
            
# linear blackbody context generator: keeps a running score of the relevant blackbodies, 
# emits a LinBbContext from time to time
class LinBbContextGenerator:
    def __init__(self, cxtType):
        self.PASSES = 2 # The instrument does two passes for each direction for each blackbody
        self.cxtType = cxtType
        
        self.rawS = [] # raw spectra appended until enough passes are completed to average down
        self.rawT = [] # ditto for times
        
        self.beginS = 0
        self.beginT = 0        
        self.endS = 0
        self.endT = 0
        
    # now it's the responsibility of this methods's caller to check the return, and if
    # nonzero, to update the linear context for this mode.
    def processBb(self, spectrum, time):
        # append new data to raw arrays
        self.rawS.append(spectrum)
        self.rawT.append(time)
        
        if len(self.rawS) < self.PASSES:
            return 0 # not enough data gathered
            
        # old end is new beginning
        self.beginS = array(self.endS)
        self.beginT = self.endT
        
        # new end is averaged from the raw arrays
        self.endS = reduce(lambda x,y: x+y, self.rawS) / self.PASSES
        self.endT = reduce(lambda x,y: x+y, self.rawT) / self.PASSES
        
        # and the raw buffers get cleared
        self.rawS = []
        self.rawT = []
        
        # first time through, do not be fooled by partly assembled context:
        if self.beginT == 0:
            return 0
    
        return LinBbContext(self.beginT, self.endT, self.beginS, self.endS, self.cxtType)
        
                                                                
# Calibration context manager. accepts blackbody spectra and associated metadata.
# - Forwards bb spectra to LinBbContextGenerator instances for processing into finished contexts.
# - Manages cache of created (immutable) contexts, and switches them out as needed.
# - returns relevant calibration info (synthetic spectra) for the time requested.
#
# There are two distinct sequences of spectra to keep in mind: incoming and outgoing, with a buffer
# inbetween. The incoming blackbody spectra are fed into the ContextGenerator. The incoming observations
# are buffered until such a time as they can be calibrated. The outgoing observations (those for whom
# a currently active context exists) are processed with this context. Thus, assuming that outgoing
# spectra are monotone in time, once an outgoing spectrum's time is past the endT of the current 
# context, this context may be discarded. If a new context is already waiting behind the curtains 
# (in the context queue), it becomes active. Otherwise, the outgoing spectrum must be buffered until 
# such a time as there is a new context with that observation's time falling in the context's [beginT, endT]
# interval.
# This kind of algorithm is a BUFFER-STRADDLER: issuance of output from the buffer is dependent on the input,
# An external buffer mediates the observation data until sufficient context is built up to process the contents.

class ShisCalContextManager:
    def __init__(self):
        self.contextGen = {  'FWABB': LinBbContextGenerator('FWABB'),
                        'BWABB': LinBbContextGenerator('BWABB'),
                        'FWHBB': LinBbContextGenerator('FWHBB'),
                        'BWHBB': LinBbContextGenerator('BWHBB') }
        
        self.contextQueue = {'FWABB': [], 'BWABB': [], 'FWHBB': [], 'BWHBB': []}
        self.contextIndex = {'FWABB': 0, 'BWABB': 0, 'FWHBB': 0, 'BWHBB': 0}
        
    # returns (ambient blackbody, hot blackbody) pair
    def getBbSpectra(self, time, scanDirection, obsType):
        abbKey = scanDirection + 'ABB'
        hbbKey = scanDirection + 'HBB'
        
        abbS = self.getBbSpectrum(time, abbKey)
        hbbS =  self.getBbSpectrum(time, hbbKey)
        
        if abbS == 'DISCARD' or hbbS == 'DISCARD':
            return 'DISCARD'
            
        if abbS == 'WAIT' or hbbS == 'WAIT':
            return 'WAIT'
        
        if DEBUG_ON and 0:
            ##### DEBUG OUTPUT #####
            cQ = self.contextQueue
            caI = self.contextIndex[abbKey]
            chI = self.contextIndex[hbbKey]
            print "calibrating %s%s @ %s with context abb:(%s,%s), hbb:(%s,%s)" % (scanDirection, obsType, time, 
                cQ[abbKey][caI].beginT, cQ[abbKey][caI].endT, 
                cQ[hbbKey][chI].beginT, cQ[hbbKey][chI].endT) 
            ##### TUPTUO GUBED #####

        return (abbS, hbbS)

    
    def getBbSpectrum(self, time, cxtType):
        if len(self.contextQueue[cxtType]) == 0:
            return 'WAIT'
            
        if self.contextQueue[cxtType][self.contextIndex[cxtType]].endT < time:
            if self.contextIndex[cxtType] >= len(self.contextQueue[cxtType]) - 1:
                return 'WAIT'
            self.contextIndex[cxtType] += 1
            return self.getBbSpectrum(time, cxtType) # and try again
        
        elif self.contextQueue[cxtType][self.contextIndex[cxtType]].beginT > time:
            if self.contextIndex[cxtType] == 0:
                return 'DISCARD'
            self.contextIndex[cxtType] -= 1
            return self.getBbSpectrum(time, cxtType) # and try again

        return self.contextQueue[cxtType][self.contextIndex[cxtType]].getSpectrumAt(time)
                            
    def processBb(self, spectrum, time, bbType, scanDirection):              
        calKey = scanDirection + bbType
        calCxt = self.contextGen[calKey].processBb(spectrum, time)
        if calCxt:
            self.contextQueue[calKey].append(calCxt)
            
            
# the calibration stage class, and supporting functions
class Obs:
    def __init__(self, spectrum, time, scanDirection, abbTemp, hbbTemp, obsType, obsId, valid):
        self.spectrum = spectrum
        self.time = time
        self.scanDirection = scanDirection
        self.abbTemp = abbTemp
        self.hbbTemp = hbbTemp
        self.obsType = obsType
        self.obsId = obsId
        self.valid = valid


# The beating heart: the calibration equation.
def calibrate(spectrum, abbSpec, hbbSpec, abbTemp, hbbTemp, planck):
    specRatio = (spectrum - abbSpec) / (hbbSpec - abbSpec)
    return specRatio.real * (planck.perfectBb(hbbTemp) - planck.perfectBb(abbTemp) ) + \
            planck.perfectBb(abbTemp)


# Putting it all together: The calibration stage manages the incoming data through ports,
# and calls on the methods of its context manager to update the context, and to get instant
# synthetic calibration info.
class CalibrationStage:
    def __init__(self, ports):
        self.ports = ports
        self.contextMgr = ShisCalContextManager()
        self.planck = Planck(ports.settings.wnScale)
        self.obsQueue = []
        
    def __call__(self):
        # here's where I translate the ports to something useful
        return self.runCalibrationStage(self.ports.spectrumIn,
                                        self.ports.ancillary.time,
                                        self.ports.ancillary.spectrumType,
                                        self.ports.ancillary.scanDirection,
                                        self.ports.ancillary.abbTemp,
                                        self.ports.ancillary.hbbTemp,
                                        self.ports.ancillary.obsId,
                                        self.ports.ancillary.valid)

    def processNextInLine(self): 
        if len(self.obsQueue) == 0:
            if DEBUG_ON:
                print "Done!" #DEBUG
            return 'DONE'
        
        nextInLine = self.obsQueue[0] # observation at the front of the queue
       
        cxt = self.contextMgr.getBbSpectra(nextInLine.time, nextInLine.scanDirection, nextInLine.obsType)
        if cxt == 'WAIT': 
            if DEBUG_ON:
                print "waiting to calibrate obs # %s at %s" % (nextInLine.obsId, nextInLine.time) #DEBUG
            return 'WAIT' #  - keep 'em waiting: context still being built

        nextInLine = self.obsQueue.pop(0)
        

        if not nextInLine.valid or nextInLine.obsType not in ('OBS','ABB','HBB','SKY'):
            if DEBUG_ON:
                print "invalid obs - returned zeros for obs # %s at %s" % (nextInLine.obsId, nextInLine.time) #DEBUG
            return zeros(len(nextInLine.spectrum))
            
        elif cxt == 'DISCARD':
            if DEBUG_ON:
                print "discard obs - returned zeros for obs # %s at %s" % (nextInLine.obsId, nextInLine.time) #DEBUG
            return zeros(len(nextInLine.spectrum))
            
        else:
            if DEBUG_ON:
                print "calibrated - returning result for obs # %s at %s" % (nextInLine.obsId, nextInLine.time) #DEBUG
            return calibrate(nextInLine.spectrum, 
                            cxt[0], cxt[1], nextInLine.abbTemp, nextInLine.hbbTemp, self.planck)

    def runCalibrationStage(self, spectrum, time, obsType, scanDirection, abbTemp, hbbTemp, obsId, valid):
        if obsType in ('ABB', 'HBB'):
            self.contextMgr.processBb(spectrum, time, obsType, scanDirection) # FIXME - add abb, hbb temp

        # append current incoming observation together with its metadata to the end of the queue
        self.obsQueue.append(Obs(spectrum, time, scanDirection, abbTemp, hbbTemp, obsType, obsId, valid))
    
        # regardless of what type of spectrum entered the queue, try
        # to process the next-in-line at the end of the queue
        return self.processNextInLine()

    # run this when no new obs are coming in, to flush out (calibrated) contents of the queue.
    def flushCalibrationStage(self):
        calibrated = self.processNextInLine()
        
        # no more new inputs updating context at this point:
        # if entry is waiting for calibration context, it may as well give up.
        if calibrated == 'WAIT':
            calibrated = zeros(len(self.obsQueue[0].spectrum))
            self.obsQueue.pop(0)

        return calibrated


class Ports:
    def __init__(self, settings, spectrumIn, spectrumOut, ancillary):
        self.settings = settings
        self.spectrumIn = spectrumIn
        self.ancillary = ancillary

        
class SettingsPort:
    def __init__(self, wnScale):
        self.wnScale = wnScale


class AncillaryPort:
    def __init__(self):
        self.time = 0
        self.spectrumType = '?'
        self.scanDirection = 'FW'
        self.abbTemp = 0
        self.hbbTemp = 0
        self.obsId = 0
        self.valid = 0


################ TEST ON SHIS FBF DIR ########################
SHIS_SCENE_MIRROR_MAP = { 
    63: '?',      # invalid entry
    65: 'ABB',    # ambient blackbody
    69: 'OBS',    # earth observation
    72: 'HBB',    # hot blackbody
    83: 'SKY'     # sky observation
}

SHIS_SCAN_DIRECTION_MAP = {
    66: 'BW',     # forward
    70: 'FW'      # backward
}

if GPLOT_ON:
    ##### GNUPLOT #####
    import Gnuplot 
    gplot = Gnuplot.Gnuplot()
    gplot.set_range('yrange',[0,200])
    gplot.set_range('xrange',[0,2044])
    ##### TOLPUNG #####

class FbfCal:
    def __init__(self):
        # assume it's running in the FBF directory of interest
        self.wn_fbf = FBF('WavenumberLW.real8.2044')
        
        self.ports = Ports(SettingsPort(self.wn_fbf.read(1)), 
                    zeros(2044), # spectrum in
                    0, # spectrum out
                    AncillaryPort())
        self.calStage = CalibrationStage(self.ports)
                    
        self.specIn_fbf = FBF('SpectrumLW_NLC.complex8.2044')
        self.specOut = [] # fill this array, then write it out at the end.
        
        self.time_fbf = FBF('epoch.real8')   # TODO - for now...
        self.mirrorPosition_fbf = FBF('sceneMirrorPosition.real4')
        self.scanDirection_fbf = FBF('scanDirection.real4')
        
        # TODO - do I use these, or a weighted average of apex, top, bottom?
        self.abbTemp_fbf = FBF('ABBapexTemp.real4')
        self.hbbTemp_fbf = FBF('HBBapexTemp.real4') 
        
        self.validMask_fbf = FBF('nlappLW.sta')  # TODO - check if this is correct

        self.fbf_index = 1
        self.fbf_max_index = self.validMask_fbf.length()
        
    def processNext(self):
        self.ports.ancillary.obsId = self.fbf_index
        self.ports.ancillary.valid = self.validMask_fbf.read(self.fbf_index)

        # protect downstream calibration code from adverse effects of bad records.
        if self.ports.ancillary.valid == 1:
            self.ports.ancillary.spectrumType = SHIS_SCENE_MIRROR_MAP[
                    int(self.mirrorPosition_fbf.read(self.fbf_index)[0])]
            self.ports.ancillary.scanDirection = SHIS_SCAN_DIRECTION_MAP[
                    int(self.scanDirection_fbf.read(self.fbf_index)[0])]
            
            self.ports.spectrumIn = self.specIn_fbf.read(self.fbf_index)
            
            self.ports.ancillary.abbTemp = self.abbTemp_fbf.read(self.fbf_index)[0]
            self.ports.ancillary.hbbTemp = self.hbbTemp_fbf.read(self.fbf_index)[0]
        
            self.ports.ancillary.time = self.time_fbf.read(self.fbf_index)[0]

        else:
            self.ports.ancillary.spectrumType = '?'
            self.ports.ancillary.scanDirection = 'FW'
       
        calibrated = self.calStage() # invoke the calibration operator
        if not calibrated == 'WAIT':
            self.specOut.append(calibrated)
            if GPLOT_ON:
                gplot.plot(calibrated) #invoke GNUPLOT
                
    def processAll(self, upto = None):
        max_index = self.fbf_max_index
        if upto:
            max_index = upto

        while(self.fbf_index < max_index):
            self.processNext()
            self.fbf_index += 1

        queued = 1
        while(queued):
            calibrated = self.calStage.flushCalibrationStage()
            if calibrated == 'DONE':
                queued = 0
            else:
                self.specOut.append(calibrated)

        out_fbf = FBF('ProtoCalOut.real8.2044')
        out_fbf.create()
        out_fbf.block_write(1, array(self.specOut))
        out_fbf.close()
        
if __name__ == '__main__':
    import sys
    fbfcal = FbfCal()
<<<<<<< CalProt.py
    fbfcal.processAll(20000)
    
=======
    if len(sys.argv) < 2:
        fbfcal.processAll()
    else:
        fbfcal.processAll(int(sys.argv[1]))    
>>>>>>> 1.13
        
