/**
 * \file CalibrationFrame.cc
 *
 * \brief Implementation of the calibration frame idea.
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 *  \version $Id: CalibrationFrame.cc,v 1.8 2005/03/11 23:12:11 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

#include <blitz/array.h>
#include <boost/shared_ptr.hpp>
#include <vector>
#include <numeric>
#include <iostream>

#include "Instrument.h"
#include "Physics.h"
#include "CalibrationFrame.h"

using namespace gips;
using physics::seconds_t;
using std::ostream;

ostream& gips::operator<<(ostream& os, const Frame& f) { return f.toStream(os); }

Frame::time_compare_t LinBbRadianceFrame::compareTime(seconds_t now) const
{
    if (now < beginT)
	return BEFORE_FRAME;
    if (now > endT)
	return AFTER_FRAME;
    return IN_FRAME;
};

void LinBbRadianceFrame::getSpectrumAt(seconds_t now,  
				       blitz::Array<complex<double>,1> &rad) const
{
    assert(compareTime(now) == IN_FRAME);
    rad = beginS + D * static_cast<double>(now - beginT);
}

ostream& LinBbRadianceFrame::toStream(ostream &os) const
{
    return os << frameId;
}



FramePtr Lin4BbFrameGenerator::getFrame()
{
    std::ostringstream oss;
    oss << frameGeneratorId << frameId++;
    return FramePtr(new LinBbRadianceFrame(beginTime, endTime, beginSpectrum, endSpectrum, oss.str()));
};

bool Lin4BbFrameGenerator::processSpectrum(const blitz::Array<complex<double>,1> &rad, seconds_t now)
{
    //std::cerr << "C++: in Lin4BbFrameGenerator::processSpectrum" << std::endl;

    timeBuffer.push_back(now);
    spectrumBuffer.push_back(rad.copy());
    
    // gathering spectra: not yet ready to emit   
    if (spectrumBuffer.size() < PASSES) 
        return false;

    assert(spectrumBuffer.size() == PASSES && timeBuffer.size() == PASSES);

    // new beginning is old end
    beginSpectrum = endSpectrum.copy();
    beginTime = endTime;
    
//    if(PASSES == 1)
//    {
//        endSpectrum = spectrumBuffer[0];
//        endTime = timeBuffer[0];
//    } 
//    else 
//    {   
    //new end is averaged from the buffer vectors
        endSpectrum = accumulate(spectrumBuffer.begin(), spectrumBuffer.end(), zeroSpectrum.copy()) 
    	    / static_cast<double>(PASSES);
        endTime = accumulate(timeBuffer.begin(), timeBuffer.end(), static_cast<seconds_t>(0)) 
    	    / static_cast<seconds_t>(PASSES);
//    }
    
    spectrumBuffer.clear();
    timeBuffer.clear();
    assert(spectrumBuffer.size() == 0 && timeBuffer.size() == 0);

    // first time through, don't have fully assembled context.
    if(beginTime == static_cast<seconds_t>(0))
	return false;

    //std::cerr << "C++: processSpectrum ready to return context" << std::endl;
    return true;
};

