/**
* \file CalibrationFrame.h
 *
 * \brief Defines a calibration frame - that is, an immutable object holding a calibration model
 * valid for a given interval of time for a given sensor, etc. Corresponds to "context" in the python 
 * prototype.
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 *  \version $Id: CalibrationFrame.h,v 1.15 2005/03/16 00:04:03 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

#ifndef H_CALFRAME
#define H_CALFRAME

#include <vector>
#include <blitz/array.h>
#include <boost/shared_ptr.hpp>

#include <iosfwd>

#include "Instrument.h"
#include "Physics.h"


namespace gips 
{
  
    class Frame
    {
    public:
	enum time_compare_t {BEFORE_FRAME, AFTER_FRAME, IN_FRAME};

	virtual ~Frame() {};
	virtual time_compare_t compareTime(physics::seconds_t now) const = 0;

	virtual void getSpectrumAt(physics::seconds_t now,  
			   blitz::Array<complex<double>,1> &rad) const = 0;

	virtual std::ostream& toStream(std::ostream &os) const = 0;
    };

    typedef boost::shared_ptr<Frame> FramePtr;


    class LinBbRadianceFrame: public Frame
    {
	const physics::seconds_t beginT;
	const physics::seconds_t endT;
	const blitz::Array<complex<double>,1> beginS;
	const blitz::Array<complex<double>,1> endS;
	const blitz::Array<complex<double>,1> D;
	const string frameId;

    public:
	LinBbRadianceFrame(physics::seconds_t bT, physics::seconds_t eT, 
			   const blitz::Array<complex<double>,1> &bS,
			   const blitz::Array<complex<double>,1> &eS,
			   const string& fid): 
	  beginT(bT), endT(eT), beginS(bS.copy()), endS(eS.copy()), 
	  D((endS-beginS)/static_cast<double>(endT-beginT)),
	  frameId(fid)
	  {};
	
	virtual ~LinBbRadianceFrame() {};

	time_compare_t compareTime(physics::seconds_t now) const;

	void getSpectrumAt(physics::seconds_t now,  
			   blitz::Array<complex<double>,1> &rad) const;

	std::ostream& toStream(std::ostream &os) const;
    };
    
    std::ostream& operator<<(std::ostream&, const Frame&);

    
    class FrameGenerator
    {
    public:
	const instrument::band_t band;
	FrameGenerator(const instrument::band_t &b): band(b) 
	{};
	
	virtual ~FrameGenerator() {};
	virtual FramePtr getFrame() = 0;
	virtual bool processSpectrum(const blitz::Array<complex<double>,1> &rad, 
				     physics::seconds_t now) = 0;
    };
    
    class Lin4BbFrameGenerator: public FrameGenerator
    {
    private:
	const unsigned int PASSES; // How many times a blackbody is stared at in a row (usually 1 or 2)
	physics::seconds_t beginTime;
	physics::seconds_t endTime;
	std::vector<physics::seconds_t> timeBuffer;
	const blitz::Array<complex<double>,1> zeroSpectrum;
	blitz::Array<complex<double>,1> beginSpectrum;
	blitz::Array<complex<double>,1> endSpectrum;
	std::vector<blitz::Array<complex<double>,1> > spectrumBuffer;
	const string frameGeneratorId;
	int frameId;
	
    public:
	explicit Lin4BbFrameGenerator(const instrument::band_t& b, unsigned int pasnum, const string& fgid): 
	    FrameGenerator(b),
	    PASSES(pasnum),
	    beginTime(0), endTime(0), 
	    zeroSpectrum(band.size), beginSpectrum(band.size), endSpectrum(band.size),
	    frameGeneratorId(fgid),
	    frameId(0)
	    {};
	    
	virtual ~Lin4BbFrameGenerator() {};
	
	virtual FramePtr getFrame();
	
	virtual bool processSpectrum(const blitz::Array<complex<double>,1> &rad, physics::seconds_t now);
    };
}

#endif
