/** 
*  \file RadCal_SHISv0.h
*  \brief Scanning HIS zeroth version radiometric calibration stage. Hold on to yer hats, this
*   ride's going to be a tad rough.
*    
*    Calibration algorithm:
*    The radiometric calibration algorithm for Michelson interferometers, using two reference blackbodies
*    
*    Comments:
* This version does not yet correctly handle laser Fringe Count Errors (FCE), which have a pervasive effect
* on the complex phase of the resulting spectrum.
*
*  \author M.J.Smuga-Otto <maciek@ssec.wisc.edu>
*
*  \version $Id: RadCal_SHISv0.cc,v 1.15 2005/03/16 00:04:03 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/
#include <iterator>
#include <list>
#include <map>
#include <queue>

#include <blitz/array.h>

#include "Instrument.h"

#include "RadiometricCalibration.h"
#include "CalibrationFrame.h"
// #include "RadCalOperator.h"
#include "Planck.h"

using namespace std;
using namespace gips;

class RadCalSHIS: public RadiometricCalibration
{
private:
    enum Q_STATE {OK, WAIT, DISCARD};

    Planck planck;
    Lin4BbFrameGenerator abbFrameGen, hbbFrameGen;
    list<FramePtr> abbFrames, hbbFrames; // list of calibration frames. 
    list<FramePtr>::iterator  abbCurrentFrame, hbbCurrentFrame; // pointers into the frames list.

    queue<RadiometricCalibration::Observation> obsQ; // queue of observations

    // private buffers for holding current generated blackbody spectra.
    blitz::Array<complex<double>,1> abbS, hbbS; // syntetic (averaged) curves 
    blitz::Array<double,1> abbB, hbbB;          // planck curves

    Q_STATE getBbSpectrum(physics::seconds_t now, 
			   const list<FramePtr> &frame, 
			   list<FramePtr>::iterator &framePtr, 
			   blitz::Array<complex<double>,1> &bb);
    void tailObsQueue();
    Q_STATE headObsQueue();
    void setOutputToInvalid();
    
public:
    RadCalSHIS(RadiometricCalibration::Ports &ports): 
	RadiometricCalibration(ports),
	planck(settings.band.wavenumbers),
        
	abbFrameGen(settings.band, settings.numpasses, "ABB"),
	hbbFrameGen(settings.band, settings.numpasses, "HBB"),
	// abbFrames(), hbbFrames(),
	abbCurrentFrame(abbFrames.begin()),
	hbbCurrentFrame(hbbFrames.begin()),
	// obsQ(), 
	abbS(settings.band.size), 
	hbbS(settings.band.size),
	abbB(settings.band.size), 
	hbbB(settings.band.size)

	{};
	
    virtual ~RadCalSHIS() {};
    virtual Operator::Result operator()();

};

RadCalSHIS::Q_STATE RadCalSHIS::getBbSpectrum(physics::seconds_t now, 
				   const list<FramePtr> &frames, 
				   list<FramePtr>::iterator &currentFrame, 
				   blitz::Array<complex<double>,1> &bb)
{
    if (frames.empty())
	return WAIT;

    Frame::time_compare_t tc(Frame::BEFORE_FRAME);
    do
    {
	if(currentFrame == frames.end())
	    currentFrame--;

	tc = (*currentFrame)->compareTime(now);
	//std::cerr << "C++: inRange returned " << tc << std::endl; // DEBUG

	if (tc == Frame::AFTER_FRAME)
	{
	    ++currentFrame;
	    
	    if (currentFrame == frames.end())
		return WAIT;
	}
	else if (tc == Frame::BEFORE_FRAME)
	{
	    if (currentFrame == frames.begin())
		return DISCARD;
	    
	    --currentFrame;
	}
    } 
    while (tc != Frame::IN_FRAME);

    (*currentFrame)->getSpectrumAt(now, bb);
    return OK;
};


Operator::Result RadCalSHIS::operator()()
{
    tailObsQueue();
    headObsQueue();
    return Operator::OK;
};


// First take care of the new observation, to be pushed to the end of the queue.
// If it's a blackbody, throw it at the appropriate context generator
void RadCalSHIS::tailObsQueue()
{
    const AncillaryObsData &ancillaryObs(ports.ancillaryObs.getForInput());
    const AncillaryBbData &ancillaryBb(ports.ancillaryBb.getForInput());

    RadiometricCalibration::Observation obs(ancillaryObs, 
					    ancillaryBb, 
					    ports.uncalibratedSpectrum.getForInput());
    obsQ.push(obs);
    
    //std::cerr << "C++: obsType at tail of queue: " << obs.ancillaryObs.obsType << std::endl; // DEBUG

    if (instrument::AMBIENT == obs.ancillaryObs.obsType){
	std::cerr << "C++: processing ambient BB" << std::endl;
	bool frameReady(abbFrameGen.processSpectrum(obs.rad, obs.ancillaryObs.obsTime));
	if(frameReady)
	{
	    abbFrames.push_back(abbFrameGen.getFrame());
	    abbCurrentFrame = abbFrames.end();
	}
    }
    else if (instrument::HOT == obs.ancillaryObs.obsType)
    {
	std::cerr << "C++: processing hot BB" << std::endl;
	bool frameReady(hbbFrameGen.processSpectrum(obs.rad, obs.ancillaryObs.obsTime));
	if(frameReady)
	{
	    hbbFrames.push_back(hbbFrameGen.getFrame());
	    hbbCurrentFrame = hbbFrames.end();
	}
    }

};


 // next, see if the observation at the front of the queue is ready to be processed.
 // if it is, run the calibrator and pop the observation out.
RadCalSHIS::Q_STATE RadCalSHIS::headObsQueue()
{   
    RadiometricCalibration::Observation &obs(obsQ.front());
    RadiometricCalibration::AuditData& au(ports.audit.getForOutput());

    //std::cerr << "C++: at head of ObsQueue for task " << obs.ancillaryObs.taskId << std::endl;
    
    au.completedTaskId = obs.ancillaryObs.taskId;

    Q_STATE hRes = getBbSpectrum(obs.ancillaryObs.obsTime, hbbFrames, hbbCurrentFrame, hbbS); // hot
    Q_STATE aRes = getBbSpectrum(obs.ancillaryObs.obsTime, abbFrames, abbCurrentFrame, abbS); // ambient
    
    if (DISCARD == hRes || DISCARD == aRes)
    {
	std::cerr << "C++: DISCARDing observation" << endl;
	obsQ.pop();
	setOutputToInvalid();
	return DISCARD; 
    }
    else if (WAIT == hRes || WAIT == aRes)
    {
	std::cerr << "C++: WAITing for calibration frame" << endl;
	setOutputToInvalid();
	au.completedTaskId = "0";
	return WAIT;
    }
    else // all's good - have valid generated blackbody spectra. Proceed with calibration.
    {
	std::cerr << "C++: processing spectrum " << au.completedTaskId 
		  << " with calibration frames"
		  << " hbb:" << **hbbCurrentFrame 
		  << " abb:" << **abbCurrentFrame <<endl;
	// FIXME - these aren't the exact computations needed here...
	double abbT = (obs.ancillaryBb.abbApexTemp
		       + obs.ancillaryBb.abbTopTemp
		       + obs.ancillaryBb.abbBottomTemp) / 3.0;
	double hbbT = (obs.ancillaryBb.hbbApexTemp
		       + obs.ancillaryBb.hbbTopTemp
		       + obs.ancillaryBb.hbbBottomTemp) / 3.0;

	planck.B(abbT, abbB);
	planck.B(hbbT, hbbB);

	//std::cerr << "C++: computed planck radiances for hbb, abb" << std::endl;

	blitz::Array<complex<double>,1>& diagS(ports.diagnosticSpectrum.getForOutput());
	blitz::Array<double,1>& outS(ports.calibratedSpectrum.getForOutput());

	//std::cerr << "obs " << obs.rad(600);
	//std::cerr << ", hbbS " << hbbS(600);
	//std::cerr << ", abbS " << abbS(600);
	//std::cerr << ", hbbB " << hbbB(600);
	//std::cerr << ", abbB " << abbB(600);

	//Operator::Result calRes = RadiometricCalibrationOperator(obs.rad,
	//							 hbbS, abbS,
	//							 hbbB, abbB,
	//							 diagS, outS);
	//assert(calRes == Operator::OK);	

	// the calibration equation
	diagS = (obs.rad - abbS) / (hbbS - abbS);
	outS = real(diagS) * (hbbB - abbB) + abbB;

	//std::cerr << ", diag " << diagS(600);
	//std::cerr << ", out " << outS(600) << std::endl;

	obsQ.pop();


	return OK;
    }

};

void RadCalSHIS::setOutputToInvalid()
{
    ports.calibratedSpectrum.getForOutput() *= 0.0;
    // passing along the unprocessed spectrum as a diagnostic
    // feels like the correct thing to do in this instance.
    ports.diagnosticSpectrum.getForOutput() = ports.uncalibratedSpectrum.getForInput(); 
};


boost::shared_ptr<gips::RadiometricCalibration> gips::RadiometricCalibrationFactory(
						      instrument::PipelineSelector sel, 
						      RadiometricCalibration::Ports &ports) 
{
    return boost::shared_ptr<gips::RadiometricCalibration>(new RadCalSHIS(ports));
};
