/** 
*  \file RadCal_SHISv0Test.h
*  \brief Unit test of Scanning HIS zeroth version radiometric calibration stage. 
*    
*  \author M.J.Smuga-Otto <maciek@ssec.wisc.edu>
*
*  \version $Id: RadCal_SHISv0Test.cc,v 1.9 2005/03/16 00:04:03 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/
#include <map>
#include <vector>
#include <iostream>

#include <blitz/array.h>
#include <boost/shared_ptr.hpp>

#include "Instrument.h"
#include "Connector.h"

#include "RadiometricCalibration.h"

using namespace std;
using namespace gips;

int main()
{
    // the self-contained test...
    
    // generate test environment
    const int SIZE(100);
    
    // settings port
    blitz::Array<double,1> WN(SIZE);
    for (int i = 0; i < SIZE; ++i)
        WN(i) = i;
    const int TEST_FOLD_POINT = 45;
    const instrument::band_t TEST_BAND(WN, TEST_FOLD_POINT);
    const unsigned int NUM_PASSES = 2;

    RadiometricCalibration::Settings settings(TEST_BAND, NUM_PASSES);
    Connector::ConstantInputPort<RadiometricCalibration::Settings> settingsPort(settings);

    // input, output and diagnostic ports
    blitz::Array<complex<double>,1> inS(SIZE);
    Connector::BlitzOpenConnector<complex<double> > dataInPort(inS);

    blitz::Array<double,1> outS(SIZE);
    Connector::BlitzOpenConnector<double> dataOutPort(outS);

    blitz::Array<complex<double>,1> diagS(SIZE);
    Connector::BlitzOpenConnector<complex<double> > dataDiagPort(diagS);

    // ancillary ports
    RadiometricCalibration::AncillaryBbData ancilBb;
    Connector::ConstantInputPort<RadiometricCalibration::AncillaryBbData> ancilBbPort(ancilBb);

    RadiometricCalibration::AncillaryObsData ancilObs;
    Connector::ConstantInputPort<RadiometricCalibration::AncillaryObsData> ancilObsPort(ancilObs);

    // audit port
    RadiometricCalibration::AuditData audit;
    Connector::SimpleConnector<RadiometricCalibration::AuditData> auditPort(audit);
    
    // gather the above ports into a single Ports structure
    RadiometricCalibration::Ports ports(settingsPort, dataInPort, dataOutPort, dataDiagPort,
					ancilObsPort, ancilBbPort, auditPort);
    
    // and construct a stage using the factory method
    boost::shared_ptr<RadiometricCalibration> stage(RadiometricCalibrationFactory(instrument::SHIS_V0, ports)); 
    
    Operator::Result res = (*stage)();
    assert(res == Operator::OK);

    //cout << outS << endl;
    return 0;
};
