/** 
*  \file RadiometricCalibration.h
*  \brief Interface specification for (linear) Radiometric Calibration Algorithm.
*    
*    Calibration algorithm:
*    The radiometric calibration algorithm for Michelson interferometers, using two reference blackbodies
*    
*    Comments:
* This version does not yet correctly handle laser Fringe Count Errors (FCE), which have a pervasive effect
* on the complex phase of the resulting spectrum.
*
*  \author M.J.Smuga-Otto <maciek@ssec.wisc.edu>
*
*  \version $Id: RadiometricCalibration.h,v 1.21 2005/03/16 00:04:03 maciek Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*  \endverbatim
*/

#ifndef H_RADCAL
#define H_RADCAL

#ifndef NO_DEPENDENCY_INCLUDES

#include <complex>

#include <blitz/array.h>
#include <boost/shared_ptr.hpp>

#include "Physics.h"
#include "Instrument.h"
#include "Operator.h"
#include "ProtocolRole.h"

#endif // #ifndef NO_DEPENDENCY_INCLUDES

namespace gips // Generalized Interferometer Pipeline Software
{

class RadiometricCalibration 
{
public:
    struct Settings
    { 
        instrument::band_t band;
	unsigned int numpasses;
	Settings(const instrument::band_t &b, unsigned int np): band(b), numpasses(np) {};
    }; // struct Settings

    // (band_t should be a struct { unsigned size, vector<double> wavenumbers }

    /**
    * \struct AncillaryData
     * \brief Immediate ancillary data needed for calculation of calibration coefficients.
     */
    struct AncillaryBbData
    {
        double hbbApexTemp; 
        double hbbTopTemp;
        double hbbBottomTemp;

        double abbApexTemp;
        double abbTopTemp;
        double abbBottomTemp;
	
        AncillaryBbData() 
	{
	    hbbApexTemp = hbbTopTemp = hbbBottomTemp =
		abbApexTemp = abbTopTemp = abbBottomTemp = 0.0;
	};
    };
    typedef boost::shared_ptr<AncillaryBbData> AncillaryBbDataPtr;

    struct AncillaryObsData
    {
	string taskId;
        physics::seconds_t obsTime;
        instrument::ObservationType obsType;
	
	AncillaryObsData():
	    taskId("0"),
	    obsTime(static_cast<physics::seconds_t>(0)),
	    obsType(instrument::INVALID)
	{};
    };
    typedef boost::shared_ptr<AncillaryObsData> AncillaryObsDataPtr;

    struct AuditData
    {
	string completedTaskId;
	// FIXME - and plenty other information coming soon!
	AuditData(): completedTaskId("0") 
	    {}
    };

    struct Observation
    {
	// taskId is recorded in ancillaryObs
	const AncillaryObsData ancillaryObs;
	const AncillaryBbData ancillaryBb;
	const blitz::Array<std::complex<double>,1> rad;
	Observation(const AncillaryObsData& aO, 
		    const AncillaryBbData& aB, 
		    const blitz::Array<std::complex<double>,1>& r):
	    ancillaryObs(aO), ancillaryBb(aB), rad(r.copy()) {};
    };
    
    
    struct Ports
    {
	// settings port
	Connector::InputPort<Settings>& settings;
	
	// real-time data input
	Connector::BlitzInputPort<std::complex<double> >& uncalibratedSpectrum;
	// real-time data output
	Connector::BlitzOutputPort<double>& calibratedSpectrum;
	Connector::BlitzOutputPort<std::complex<double> >& diagnosticSpectrum; 
	// NOTE: diagnostic output is complex in this version,
	// to enable easy hook-up to alternate calibration/analysis stages.
	
	// observation ancillary data
	Connector::InputPort<AncillaryObsData>& ancillaryObs;
	// blackbody ancillary data
	Connector::InputPort<AncillaryBbData>& ancillaryBb;

	// audit data.
	Connector::OutputPort<AuditData>& audit;
	
	Ports(Connector::InputPort<Settings>& s,
	      Connector::BlitzInputPort<std::complex<double> >& uS,
	      Connector::BlitzOutputPort<double>& cS,
	      Connector::BlitzOutputPort<std::complex<double> >& dS,
	      Connector::InputPort<AncillaryObsData>& aO,
	      Connector::InputPort<AncillaryBbData>& aB,
	      Connector::OutputPort<AuditData>& au):
	    settings(s), uncalibratedSpectrum(uS), calibratedSpectrum(cS), 
	    diagnosticSpectrum(dS), ancillaryObs(aO), ancillaryBb(aB), audit(au) 
	    {};
	
    }; // struct Ports
    
protected:
    Ports &ports;
    const Settings settings;
    
    explicit RadiometricCalibration(Ports &);
     virtual ~RadiometricCalibration() = 0;
     
public:
     virtual Operator::Result operator()() = 0;
     
}; // end class RadiometricCalibration

boost::shared_ptr<RadiometricCalibration> RadiometricCalibrationFactory(instrument::PipelineSelector, 
						      RadiometricCalibration::Ports &);
}; // end namespace gips


#endif // #ifndef H_RADCAL

