/**
 * \file Ifg2Spectrum.h
 * \brief declarations (not definitions) for the Interferogram to Spectrum abstract stage
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: Ifg2Spectrum.h,v 1.19 2005/03/08 22:53:41 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 *
 */

/**
 * \mainpage Converting Interferograms to Spectra
 *
 * \image html overview.jpg
 *
 * \image html capsule_diagram.jpg
 *
 * \section TOO Theory of Operation
 * TBD
 * \section IN Implementation Notes
 * TBD
 *   
 */

#ifndef H_IFG2SPEC_STAGE
#define H_IFG2SPEC_STAGE

#ifndef NO_DEPENDENCY_INCLUDES
// GIPS includes
#include "Physics.h"
#include "Instrument.h"
#include "ProtocolRole.h"
#include "Operator.h"

#include <boost/shared_ptr.hpp>

#endif

using namespace gips;

namespace Ifg2Spectrum
{
    /// Configuration (initialization time) settings for the Inteferogram to Spectrum operator. 
    struct Config {
        size_t npts; ///< number of points in input interferogram (and output spectrum)
        int foldPoint; ///< point around which transformed spectrum is folded. See SpecFoldOperator.
        Config(const size_t& n, const int& f ): npts( n ), foldPoint( f ) { }; ///< ctor
    };
        
    typedef  Connector::BlitzInputPort<std::complex<double> > JackIn;
    typedef  Connector::BlitzOutputPort<std::complex<double> > JackOut;
    typedef  Connector::InputPort<Config> JackConfig;

    
/// Ports for the Interferogram to Spectrum stage
struct Ports {
    JackIn& inputInterferogram; ///< raw interferogram input
    JackOut& outputSpectrum; ///< raw spectrum output
    JackConfig& config; ///< configuration settings
    
    Ports(JackIn&, JackOut&, JackConfig&); ///< ctor
};
        
    
/// \brief This stage of the L0-L1 pipeline transforms a complex interferogram 
/// (such as might come off the raw data stream from the instrument) into a complex
/// spectrum.
class Ifg2SpectrumStage: Operator::AbstractOperator
{
protected:    
    Ports &ports;
    explicit Ifg2SpectrumStage(Ports&);
public:
    virtual Operator::Result operator()( ); ///< make it so
    virtual ~Ifg2SpectrumStage(); // abstract class

    /// factory method which produces instances of an appropriate subclass for the given implementation.
    static boost::shared_ptr<Ifg2SpectrumStage> Factory( instrument::PipelineSelector, Ports & );
}; // class Ifg2SpectrumStage

} // namespace
#endif

