/*
 * \brief SHIS implementation of Ifg2SpectrumStage
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: SHIS_Ifg2Spectrum.cc,v 1.21 2005/03/08 22:53:41 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */


#include <complex>
#include <vector>

#include <boost/shared_ptr.hpp>

#include "Ifg2Spectrum.h"

#include "Connector.h"
#include "IfgShift.h"
#include "CplxFFT.h"
#include "SpectrumFold.h"


using namespace gips;

namespace Ifg2Spectrum
{

// a shim to distribute the external configuration connector among the member operators 
struct ConfigDistributor
{
    const Config &inConfig;
    
    IfgShift::Config ifgShiftConfig;
    CplxFFT::Config  cplxFFTConfig;
    SpecFold::Config specFoldConfig;
    
    Connector::ConstantInputPort<IfgShift::Config> shiftConfigJack;
    Connector::ConstantInputPort<CplxFFT::Config> fftConfigJack;
    Connector::ConstantInputPort<SpecFold::Config> foldConfigJack;
    
    ConfigDistributor( Connector::InputPort<Config> &configConnector ):
        inConfig( configConnector.getForInput() ), 
        ifgShiftConfig( inConfig.npts ), 
        cplxFFTConfig( inConfig.npts ),
        specFoldConfig( inConfig.npts, inConfig.foldPoint ),
        shiftConfigJack( ifgShiftConfig),
        fftConfigJack( cplxFFTConfig ),
        foldConfigJack( specFoldConfig )
    { }; // ctor
        
}; // struct ConfigDistributor
    
class SHISStage: public Ifg2SpectrumStage
{
private:
    // create distributor for configuration connector (external to internal)
    ConfigDistributor configDist;
    
    // create intermediate buffers as connectors
    Connector::BlitzClosedConnector< std::complex<double> > shiftedIfg;
    Connector::BlitzClosedConnector< std::complex<double> > prefoldedSpectrum;

    
    // and set up the operator classes using these connectors
    IfgShift::Ports shiftPorts;
    IfgShift::IfgShiftOperator ifgShift;
    
    CplxFFT::Ports fftPorts;
    CplxFFT::CplxFFTOperator cplxFFT;
    
    SpecFold::Ports foldPorts;
    SpecFold::SpecFoldOperator foldSpectrum;
public:
    explicit SHISStage( Ports &ports ): Ifg2SpectrumStage( ports ),
        configDist( ports.config ),
        shiftedIfg(configDist.inConfig.npts), 
        prefoldedSpectrum(configDist.inConfig.npts),
        shiftPorts( ports.inputInterferogram, shiftedIfg, configDist.shiftConfigJack ),
        ifgShift( shiftPorts),
        fftPorts( shiftedIfg, prefoldedSpectrum, configDist.fftConfigJack ),
        cplxFFT( fftPorts ),
        foldPorts( prefoldedSpectrum, ports.outputSpectrum, configDist.foldConfigJack ),
        foldSpectrum( foldPorts )
    {};  
    
    virtual ~SHISStage() {};
    virtual Operator::Result operator() ( );
}; // class Stage
    
Operator::Result SHISStage::operator() ( )
{
    // call the operators in sequence
    //std::cerr << "C++: in Ifg2Spectrum::operator()" << std::endl;
    //const instrument::BlitzCplxDoubleVector ifgPayload( shiftedIfg.getForInput() ) ; // for debugging purposes
    //const instrument::BlitzCplxDoubleVector specPayload( prefoldedSpectrum.getForInput() ); // for debugging purposes

    int res = 0;
    //std::cerr << "C++: about to shift Ifg" << std::endl;
    res |= (int)ifgShift();
    //std::cerr << "ifgPayload channels 0,10,200: " << ifgPayload(0) << "," << ifgPayload(10) << "," << ifgPayload(200) << std::endl;
    res |= (int)cplxFFT();
    //std::cerr << "specPayload channels 0,10,200: " << specPayload(0) << "," << specPayload(10) << "," << specPayload(200) << std::endl;
    res |= (int)foldSpectrum();
        
    return res ? Operator::ERROR : Operator::OK;
} // operator()
    
boost::shared_ptr<Ifg2SpectrumStage> Ifg2SpectrumStage::Factory(instrument::PipelineSelector pipeline, Ports &p)
{
        return boost::shared_ptr<Ifg2SpectrumStage>(new SHISStage(p));
} // Factory
    
} // namespace

