/**
 * \file L0l1Pipeline.h
 * \brief Abstract capsule interface for the Level 0 (L0) to Level 1 (L1) science pipeline
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 *  \version $Id: L0L1Pipeline.h,v 1.7 2005/03/16 00:04:03 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2005
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 *
 */

/**
* \mainpage Converting raw interferograms to calibrated spectra.
 *
 * \image html overview.jpg
 *
 * \image html capsule_diagram.jpg
 *
 * TODO - link to these to correct image files.   
 */

#ifndef H_L0L1_PIPELINE
#define H_L0L1_PIPELINE

#ifndef NO_DEPENDENCY_INCLUDES

#include "boost/shared_ptr.hpp"

// generic GIPS includes
#include "Physics.h"
#include "Instrument.h"
#include "ProtocolRole.h"
#include "Operator.h"

#endif

using std::complex;

namespace gips
{
    class L0L1Pipeline
    {
    public:
	struct Settings {
	    instrument::band_t band; ///< SW, MW or LW
	    unsigned int blackbodyPasses;
	    Settings(const instrument::band_t &b, const unsigned int bp): band(b), blackbodyPasses(bp) {};
	};

	struct Ancillary {
	    // all the metadata fit to accompany an observation.
	    string taskId;
	    physics::seconds_t time;
	    instrument::pixel_t pixel;
	    instrument::ScanDirection scanDirection;
	    instrument::ObservationType obsType;
	    Ancillary(): taskId("0"), time(0), 
			 pixel(static_cast<instrument::pixel_t>(0)),
			 scanDirection(instrument::FW),
			 obsType(instrument::INVALID) {};
	};
		
	struct SpectralCalAncillary {
	    // ancillary data needed for constructing spectral calibration contexts,
	    // such as choice of ILS convolution kernels or neon reference fringe comparisons.
	    // updated on the order of hours or coarser.
	    SpectralCalAncillary() {};
	};
		    
	struct RadiometricCalAncillary {
	    // ancillary data needed for constructing radiometric calibration contexts,
	    // such as blackbody thermistor counts, etc...
	    // usually accompanies either all or at the very least blackbody spectra.
	    float abbTopTemp;
	    float abbApexTemp;
	    float abbBottomTemp;
	    
	    float hbbTopTemp;
	    float hbbApexTemp;
	    float hbbBottomTemp;	 
	    
	    RadiometricCalAncillary()
	    {
		abbTopTemp = abbApexTemp = abbBottomTemp = 
		    hbbTopTemp = hbbApexTemp = hbbBottomTemp = 0.0;
	    };
	    
	};
	
	struct Audit {
	    // data recorded about processing. Contains a trace of the processing history
	    // FIXME - will contain plenty more data besides just the record number
	    string completedTaskId;

	    Audit(): completedTaskId("0") {};
	};
	
	// data channel
	typedef Connector::BlitzInputPort<complex<double> > IfgIn;
	typedef Connector::BlitzOutputPort<double> SpecOut;
	typedef Connector::BlitzOutputPort<complex<double> > DiagOut;
	
	// other inputs 
	typedef Connector::InputPort<Settings> SettingsIn;
	typedef Connector::InputPort<Ancillary> AncilIn;
	typedef Connector::InputPort<SpectralCalAncillary> SpecAncilIn;
	typedef Connector::InputPort<RadiometricCalAncillary> RadAncilIn;
	
	// other outputs
	typedef Connector::OutputPort<Audit> AuditOut;
	
	
	struct Ports {
	    IfgIn& interferogramIn;
	    SpecOut& spectrumOut;
	    
	    SettingsIn& settings;
	    AncilIn& ancillary;
	    SpecAncilIn& spectralAncil;
	    RadAncilIn& radiometricAncil;
	   
	    DiagOut& diagnosticOut;
	    AuditOut& auditRecordsOut;
	    
	    Ports(IfgIn& i, SpecOut& o, SettingsIn& e, AncilIn& a, SpecAncilIn& s, RadAncilIn& r, DiagOut& d, AuditOut& t):
		interferogramIn(i), spectrumOut(o), settings(e), ancillary(a), spectralAncil(s), 
		radiometricAncil(r), diagnosticOut(d), auditRecordsOut(t) {};

	}; // struct Ports
	
    protected:
	Ports &ports;
	explicit L0L1Pipeline(Ports&);
    public:
	virtual Operator::Result operator()();
	virtual ~L0L1Pipeline();
	
	static boost::shared_ptr<L0L1Pipeline> Factory(const instrument::PipelineSelector, Ports&);
    
    }; // class 
    
}; // namespace

#endif
    
