/**
 * \file PipelineAdapter.cc
 * \brief unit test for SHIS pipeline.
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * C++ end of UW-SSEC pipeline runtime environment. vector data are assumed passed in and out
 * via memory maps, while metadata are passed in and out via stdin/stdout.
 *
 *  \version $Id: L0L1PipelineMmapAdapter_SHISv0.cc,v 1.2 2005/03/16 00:04:03 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 *
 */

#include <cstdlib>

#include <iostream>
#include <map>
#include <string>
#include <sstream>

#include "blitz/array.h"
#include "boost/shared_ptr.hpp"
#include "boost/tokenizer.hpp"

#include "Physics.h"
#include "Instrument.h"
#include "Connector.h"
#include "MmapWrap.h"
#include "L0L1Pipeline.h"

using namespace std;
using namespace gips;


// this is the adapter to the python surroundings, via memory maps and stdin/stdout

const boost::char_separator<char> kvSeparator(":");
typedef boost::tokenizer<boost::char_separator<char> > KVTokenizer;

struct TaskToken
{
    string taskId;
    physics::seconds_t time;
    instrument::ScanDirection scanDirection;
    instrument::ObservationType observationType;

    TaskToken(string id, physics::seconds_t t, instrument::ScanDirection sd, instrument::ObservationType ot):
	taskId(id), time(t), scanDirection(sd), observationType(ot)
	{};
    
    TaskToken(istringstream& ssin)
	{	    
	    //cerr << "C++ building task token" << endl; // DEBUG
	    map<string, string> inputMap;
	    string kvStr;
	    while(ssin >> kvStr)
	    {
		KVTokenizer kvt(kvStr, kvSeparator);
		string k = *kvt.begin();
		string v = *(++(kvt.begin()));
		inputMap[k] = v;
	    }
	    taskId = inputMap["TASK"];
	    time = atof(inputMap["time"].c_str());
	    scanDirection = instrument::getScanDir(inputMap["PIPELINE"]);
	    observationType = instrument::getObsType(inputMap["obsType"]);
	    //cerr << "C++ built task token with taskId " << taskId << " time " << time << " scanDirection " << scanDirection 
	    // << " and observation type " << observationType << endl; // DEBUG
	};
};


class PipelineEnv
{
    const string bandStr;                    // holds specification of mmaps (passed in via stdin)
    const int bandSize;
    MmapWrap<double> wnInMmap;             // memory map for band wavenumbers (used only at start)
    blitz::Array<double, 1> wnIn;          // blitz array of the same.
    instrument::band_t band;              // settings for pipeline
    L0L1Pipeline::Settings settings;

    L0L1Pipeline::Ancillary ancilIn; // no need for any special init - filled from stdin
    L0L1Pipeline::RadiometricCalAncillary radAncilIn; // ditto - filled from memmap
    L0L1Pipeline::SpectralCalAncillary specAncilIn; // not filled right now.

    MmapWrap<complex<double> > dataInMmap;
    blitz::Array<complex<double>, 1> dataIn; // Interferogram in, via memory map
    
    MmapWrap<double> dataOutMmap;
    blitz::Array<double, 1> dataOut;           // Spectrum Out, also via memory map

    blitz::Array<complex<double>, 1> diagOut; // Diagnostic out, currently discarded

    MmapWrap<double> tempsInMmap;                // temperatures in, via memory map

    L0L1Pipeline::Audit auditOut;

    typedef boost::shared_ptr<L0L1Pipeline> pipelinePtr;
    map<string, pipelinePtr> pipelines;  // the pipeline

    Connector::SimpleConnector<L0L1Pipeline::Settings> settingsConn;
    Connector::BlitzOpenConnector<complex<double> > dataInConn;
    
    Connector::BlitzOpenConnector<double> dataOutConn;
    Connector::BlitzOpenConnector<complex<double> > diagOutConn;
	
    Connector::SimpleConnector<L0L1Pipeline::Ancillary> ancilConn;
    Connector::SimpleConnector<L0L1Pipeline::RadiometricCalAncillary> radAncilConn;
    Connector::SimpleConnector<L0L1Pipeline::SpectralCalAncillary> specAncilConn;
    Connector::SimpleConnector<L0L1Pipeline::Audit> auditConn;
    
    L0L1Pipeline::Ports ports;

public:

    PipelineEnv():
	bandStr(getenv("GIPS_BAND")),
	bandSize(instrument::getBandSize(bandStr)),
	wnInMmap(atoi(getenv("GIPS_WN_IN_MMAP")), bandSize),
	wnIn(wnInMmap.getRawPtr(), blitz::TinyVector<int,1>(bandSize), blitz::neverDeleteData),
	band(wnIn, instrument::getBandFoldPoint(bandStr)), 
	settings(band, atoi(getenv("GIPS_BBPASSES"))),

	dataInMmap(atoi(getenv("GIPS_DATA_IN_MMAP")), bandSize),
	dataIn(dataInMmap.getRawPtr(), blitz::TinyVector<int,1>(bandSize), blitz::neverDeleteData),
	
	dataOutMmap(atoi(getenv("GIPS_DATA_OUT_MMAP")), bandSize, true),
	dataOut(dataOutMmap.getRawPtr(), blitz::TinyVector<int,1>(bandSize), blitz::neverDeleteData),

	diagOut(bandSize),
	
	tempsInMmap(atoi(getenv("GIPS_TEMPS_IN_MMAP")),6),
	
	settingsConn(settings), dataInConn(dataIn), dataOutConn(dataOut), diagOutConn(diagOut), ancilConn(ancilIn),
	radAncilConn(radAncilIn), specAncilConn(specAncilIn), auditConn(auditOut),
	ports(dataInConn, dataOutConn, settingsConn, ancilConn, specAncilConn, radAncilConn, diagOutConn, auditConn)
	{}
     
    ~PipelineEnv()
	{}

    void createPipeline(const string& pipelineKey)
	{
	    //cerr << "C++: wn from " << wnIn(0) << " to " << wnIn(bandSize - 1) << endl; // DEBUG
	    pipelines[pipelineKey] = L0L1Pipeline::Factory(instrument::SHIS_V0, ports);
	    cout << "CREATED:" << pipelineKey << endl;
	}; 
    

    string runPipeline(const string& pipelineKey, const TaskToken& tt)
    {
	pipelinePtr pipeline = pipelines[pipelineKey];
	//cerr << "C++: In runPipeline" << endl; // DEBUG
	//cerr << "C++: data in from " << dataIn(0) << " to " << dataIn(bandSize - 1) << endl; // DEBUG
	// update the structs behind the connectors...
	// bridge from taskToken to pipeline key + obsAncil data
	
	ancilIn.taskId = tt.taskId;
	ancilIn.time = tt.time;
	ancilIn.obsType = tt.observationType;
	// unpack the blackbody temperatures mmap to the penv.radAncilIn struct...
	//cerr << "C++ about to read blitzTemps" << endl; // DEBUG
	try 
	{
	    radAncilIn.abbTopTemp = tempsInMmap[0];
	    radAncilIn.abbApexTemp = tempsInMmap[1];
	    radAncilIn.abbBottomTemp = tempsInMmap[2];
	    radAncilIn.hbbTopTemp = tempsInMmap[3];
	    radAncilIn.hbbApexTemp = tempsInMmap[4];
	    radAncilIn.hbbBottomTemp = tempsInMmap[5];
	    //cerr << "C++: at:" << radAncilIn.abbTopTemp << " hb:" << radAncilIn.hbbBottomTemp << endl;

	} 
	catch(...)
	{
	    cerr << "C++: error in reading blitzTemps" << endl;
	    assert("blitzTemps Unread");
	}
	//cerr << "C++: ready to cycle pipeline operator" << endl; // DEBUG
	Operator::Result res = (*pipeline)();
	assert(res == Operator::OK);
	//cerr << "C++: pipeline returned code " << res << endl; // DEBUG
	
	cerr << "C++: pipelineAdapter out@600 " << dataOut(600,0) << endl; // DEBUG

	return "PROCESSED:" + pipelineKey + " TASK:" + auditOut.completedTaskId;
    };

};


int main()
{
    //cerr << "C++ alive and kicking" << endl; // DEBUG
    // construct PipelineEnv from supplied memory maps
    PipelineEnv penv;
    //cerr << "C++ still alive and kicking..." << endl; // DEBUG
    string commandStr;
    while(getline(cin, commandStr))
    {    
        //cerr << "C++ pipeline got:" << commandStr << endl; // DEBUG
	//boost::tokenizer<> commandChunks(commandStr);
	istringstream ssin(commandStr);
	string commandName;
	ssin >> commandName;
	  if (commandName == "CREATE")
	{
	    // creating pipeline
	    string commandChunk;
	    ssin >> commandChunk;
	    KVTokenizer kvt(commandChunk, kvSeparator);
	    // DEBUG...
	    //cerr << "C++ kv:";
	    //for (KVTokenizer::iterator kvi = kvt.begin(); kvi != kvt.end(); ++kvi) // DEBUG
	    //{ 
	    //    cerr << *kvi << "#";  // DEBUG
	    //}
	    //cerr << "$" << endl; // DEBUG

	    penv.createPipeline(*(++(kvt.begin())));
	}
	  else if (commandName == "PROCESS")
	{	    
	    // pipeline setup
	    TaskToken tt(ssin);
	    // pipelines are keyed by scan direction (separate pipelines for FW, BW
	    string ret = penv.runPipeline(instrument::scanDirStr(tt.scanDirection), tt);
	    cout << ret << endl;
	}
	else
	    assert("currently only CREATE and PROCESS commands are accepted");
    }
    return(0);
};
	


