/*
 * \brief SHIS implementation of Ifg2SpectrumStage
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * A Scanning-HIS implementation of the GIPS L0-L1 science pipeline.
 *
 *  \version $Id: L0L1Pipeline_SHISv0.cc,v 1.16 2005/03/16 00:04:03 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

#include <complex>
#include <vector>
#include <iostream>

#include "boost/shared_ptr.hpp"

#include "Ifg2Spectrum.h"
#include "RadiometricCalibration.h"

#include "Connector.h"
#include "IfgShift.h"
#include "CplxFFT.h"
#include "SpectrumFold.h"

#include "L0L1Pipeline.h"

using namespace std;
using namespace gips;

class SHISPipeline: public L0L1Pipeline
{
private:
    struct PortAdapter
    {
	const Settings& settings;
	
	const Ifg2Spectrum::Config i2sConfig;
	Connector::ConstantInputPort<Ifg2Spectrum::Config> i2sConfigConn;

	const RadiometricCalibration::Settings calSettings;
	Connector::ConstantInputPort<RadiometricCalibration::Settings> calSettingsConn;

	RadiometricCalibration::AncillaryBbData calBbAncilBuffer;
	Connector::SimpleConnector<RadiometricCalibration::AncillaryBbData> calBbAncilConn;

	RadiometricCalibration::AncillaryObsData calObsAncilBuffer;
	Connector::SimpleConnector<RadiometricCalibration::AncillaryObsData> calObsAncilConn;
	
	// provide memory backing for the internal data channel
	Connector::BlitzClosedConnector<complex<double> > bufConn; 

	RadiometricCalibration::AuditData auditBuffer;
	Connector::SimpleConnector<RadiometricCalibration::AuditData> auditConn;

	Ifg2Spectrum::Ports ifgPorts;
	RadiometricCalibration::Ports calPorts;

	PortAdapter(L0L1Pipeline::Ports &p): 
	    settings(p.settings.getForInput()),
	    
	    i2sConfig(settings.band.size, settings.band.foldPoint),
	    i2sConfigConn(i2sConfig),

	    calSettings(settings.band, settings.blackbodyPasses),
	    calSettingsConn(calSettings),

	    calBbAncilBuffer(),
	    calBbAncilConn(calBbAncilBuffer),
	    
	    calObsAncilBuffer(),
	    calObsAncilConn(calObsAncilBuffer),
	    
	    bufConn(settings.band.size),
	    
	    auditBuffer(),
	    auditConn(auditBuffer),
	   
	    ifgPorts(p.interferogramIn, bufConn, i2sConfigConn),
	    calPorts(calSettingsConn, bufConn, p.spectrumOut, p.diagnosticOut, 
		     calObsAncilConn, calBbAncilConn, auditConn) 
	    {};
	    
    } portAdapter;
    
    // current components only
    Ifg2Spectrum::Ports &isPorts;
    RadiometricCalibration::Ports &rcPorts;
    
    boost::shared_ptr<Ifg2Spectrum::Ifg2SpectrumStage> isStage;
    boost::shared_ptr<RadiometricCalibration> rcStage;
    
    void updateAncilBuffers();
    void updateAudit();

public:
    explicit SHISPipeline(Ports &p): 
	L0L1Pipeline(p), 
	portAdapter(ports), 

	isPorts(portAdapter.ifgPorts),
	rcPorts(portAdapter.calPorts),

	isStage(Ifg2Spectrum::Ifg2SpectrumStage::Factory(instrument::SHIS_V0, isPorts)),
	rcStage(RadiometricCalibrationFactory(instrument::SHIS_V0, rcPorts))
	
	{}; // constructor
     
    virtual Operator::Result operator()();
    virtual ~SHISPipeline() {};
    
}; // class SHISPipeline
    
void SHISPipeline::updateAncilBuffers()
{
    const Ancillary &inOBuf(ports.ancillary.getForInput());
    const RadiometricCalAncillary &inBBuf(ports.radiometricAncil.getForInput());

    RadiometricCalibration::AncillaryObsData &calOBuf(portAdapter.calObsAncilBuffer);
    RadiometricCalibration::AncillaryBbData &calBBuf(portAdapter.calBbAncilBuffer);

    calBBuf.hbbApexTemp = inBBuf.hbbApexTemp;
    calBBuf.hbbTopTemp = inBBuf.hbbTopTemp;
    calBBuf.hbbBottomTemp = inBBuf.hbbBottomTemp;

    calBBuf.abbApexTemp = inBBuf.abbApexTemp;
    calBBuf.abbTopTemp = inBBuf.abbTopTemp;
    calBBuf.abbBottomTemp = inBBuf.abbBottomTemp;

    calOBuf.taskId = inOBuf.taskId;
    calOBuf.obsTime = inOBuf.time;
    calOBuf.obsType = inOBuf.obsType;
};

void SHISPipeline::updateAudit()
{
    Audit &auditOut(ports.auditRecordsOut.getForOutput());
    const RadiometricCalibration::AuditData &auditBuf(portAdapter.auditBuffer);
    auditOut.completedTaskId = auditBuf.completedTaskId;
    // std::cerr << "C++: auditOut should be " <<  auditOut.completedTaskId << std::endl;
};


Operator::Result SHISPipeline::operator()()
{
    //cerr << "C++: in SHISPipeline::operator()" << endl;

    updateAncilBuffers();
    //cerr << "C++: updated ancil buffers" << endl;

    //const Ancillary &aa(ports.ancillary);
    //cerr << "C++: ancil buffer in Pipeline task:" << aa.taskId << " obsType:" << aa.obsType << endl; // DEBUG

    Operator::Result isRes = (*isStage)();
    //cerr << "C++: Ifg2Spectrum stage completed" << endl;

    Operator::Result rcRes = (*rcStage)();    
    //cerr << "C++: Radiometric Calibration stage completed" << endl;
    
    updateAudit();

    //blitz::Array<double,1>& sout(ports.spectrumOut.getForOutput());
    //cerr << "C++: output@100: " << sout(100) << endl;

    return Operator::OK == isRes && Operator::OK == rcRes 
      ? Operator::OK 
      : Operator::ERROR;
};

boost::shared_ptr<L0L1Pipeline> L0L1Pipeline::Factory(instrument::PipelineSelector s, L0L1Pipeline::Ports &p) 
{
    boost::shared_ptr<L0L1Pipeline> ret(new SHISPipeline(p));
    return ret;
};

