/*
 * \file L0L1Pipeline_SHISv0FBF.cc
 * \brief SHIS FBF flatfile based processing pipeline.
 *
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 * \version $Id: L0L1Pipeline_SHISv0FBF.cc,v 1.2 2005/02/17 07:30:59 maciek Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

// C++ STL includes
#include <string>
#include <vector>
#include <iostream>

// GIPS support library includes
#include "FBF.h"
#include "clok.h"

// GIPS framework includes
#include "Instrument.h"
#include "Connector.h"
#include "L0L1Pipeline.h"

using namespace std;
using namespace gips;

Instrument::band_t* getBandFromFBF(const string& bandName)
{
    string wnFilename;
    if (bandName == "LW"){
	return new Instrument::band_t("WavenumberLW.real8.2044", Instrument::SHIS_LW_FOLDPOINT);
    else if (bandName == "MW")
	return new Instrument::band_t("WavenumberMW.real8.4088", Instrument::SHIS_MW_FOLDPOINT);
    else (bandName == "SW")
	return new Instrument::band_t("WavenumberSW.real8.4088", Instrument::SHIS_SW_FOLDPOINT);
    else
	return NULL;
}

Operator::Result processFromSHISFiles(const int& beginRecord, const int& numRecords) 
{
    const string INTERFEROGRAM_IN_FILENAME;
    const string SPECTRUM_OUT_FILENAME;
    const string SPECTRUM_DIAGNOSTIC_FILENAME;

    const string ABB_TOP_TEMPERATURE_FILENAME;
    const string ABB_BOTTOM_TEMPERATURE_FILENAME;
    const string ABB_APEX_TEMPERATURE_FILENAME;

    const string HBB_TOP_TEMPERATURE_FILENAME;
    const string HBB_BOTTOM_TEMPERATURE_FILENAME;
    const string HBB_APEX_TEMPERATURE_FILENAME;

    const string TIME_FILENAME;
    const string OBS_TYPE_FILENAME;
    const string SCAN_DIRECTION_FILENAME;
    const string VALID_FLAG_FILENAME;

    // data channel in
    vector<complex<double> > rawInput;
    FILE *infp = fopen(INTERFEROGRAM_IN_FILENAME, "rb");
    FBF::load(rawInput, infp);
    fclose(infp);
    
    const blitz::Array< complex<double>, 2> blitzIn(&rawInput[0], 
						    blitz::shape(1000,2044), 
						    blitz::deleteDataWhenDone);
    Connector::BlitzMatrixBackedInputConnector<complex<double> > testIn(blitzIn);
    
    // data channel out
    blitz::Array<double, 2 > blitzOut(blitz::shape(1000,2044));
    Connector::BlitzMatrixBackedOutputConnector<double> testOut(blitzOut);

    blitz::Array<complex<double>, 2 > blitzDiag(blitz::shape(1000,2044));
    Connector::BlitzMatrixBackedOutputConnector<complex<double> > testDiag(blitzDiag);
    
    
    // and the test settings
    Config_t defaultConfig( 2044, 829 );
    Connector::ConstantInputPort< Config_t > testConfig( defaultConfig );
    
    // set up the run
    Ports ports( testIn, testOut, testConfig );
    Ifg2SpectrumStage *stageptr = Ifg2SpectrumStage::Factory( Instrument::SHIS_V0, ports );
    
    
    // run!
    Operator::Result res;
    Clok timer;
    std::cerr << "data loaded. starting main loop \n";
    timer.start();
    for (int i = 0; i < 1000; ++i)
    {
        res = stageptr->operator()( );
        assert (res == Operator::OK);
        testIn.next();
        testOut.next();
    }
    timer.stop();
    std::cerr << "finished processing. Time stats: " << timer.report() << "\nsaving result\n";
    
    if (res == Operator::OK) {
        // and write the results to file
        int stat = FBF::blitz_save( OUTPUT_FILENAME, blitzOut );
        if (stat != 0 )
            res = Operator::ERROR;
    }
    
    return res;
};


int main( int argc, char *argv[] )
{
    Operator::Result rc;
    
    if (argc == 3) // gave input, output filenames as arguments
    {
        rc = Ifg2Spectrum::spectrumFileTest(argv[1], argv[2]);
        assert (rc==Operator::OK);
    }
    else
    {
            rc = Ifg2Spectrum::livelinessTest();
            assert( rc==Operator::OK );
    }
        
    
    exit(0);
}
