#!/usr/bin/env python

""" Prototype wrapper for C++ GIPS pipeline, version SHISv0

This script wrapper sources the appropriate FBF files, 
and spins off the C++ PipelineAdapter as a separate process
connected via pipes that masquerade as stdin/stdout on the C++ end.
(non meta-)data is passed along via memory maps.

Author: Maciek Smuga-Otto <maciek@ssec.wisc.edu>

version $Id: PipelineWrapperSHISv0.py,v 1.18 2005/03/17 16:22:54 maciek Exp $

Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
Space Science and Engineering Center
University of Wisconsin - Madison, USA

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from FBF import FBF
from Numeric import *
from os import *

GPLOT_ON = 0 ### - set to 0 if Gnuplot is not installed, or not desired.

SHIS_BAND_SIZE = {'LW': 2044,
                  'MW': 4088,
                  'SW': 4088 
                 }

DATA_IN_SELECTOR = {
    "IFG":      "Interferogram%s.complex8.%s",
    "SPEC_RAW": "Spectrum%s.complex8.%s",
    "SPEC_NLC": "Spectrum%s_NLC.complex8.%s"
    }

def getFbfSources(band, data_in_name="IFG"):
    """ create and return FBF objects for files needed to run SHISv0 pipeline.

    All but the data_out file need to exist in the local dir. the data_out file is created."""

    class FbfSrc:
        def __init__(self, name):
            self.name = name
            self.fbf = FBF(name)

    bandsize = SHIS_BAND_SIZE[band]

    fs = {'data_in': FbfSrc(DATA_IN_SELECTOR[data_in_name] % (band, bandsize)),
          'data_out': FbfSrc("Spectrum%s_SHISv0.real8.%s" % (band, bandsize)),
          
          'at_temp': FbfSrc("ABBtopTemp.real4"),
          'ab_temp': FbfSrc("ABBbottomTemp.real4"),
          'aa_temp': FbfSrc("ABBapexTemp.real4"),
          
          'ht_temp': FbfSrc("HBBtopTemp.real4"),
          'hb_temp': FbfSrc("HBBbottomTemp.real4"),
          'ha_temp': FbfSrc("HBBapexTemp.real4"),
          
          'time': FbfSrc("epoch.real8"),
          'obsType': FbfSrc("sceneMirrorPosition.real4"),
          'scanDir': FbfSrc("scanDirection.real4"),
          'valid': FbfSrc("rsh2fbf.sta"),
          'wn_in':FbfSrc("Wavenumber%s.real8.%s" % (band, bandsize))
          }

    fs['data_out'].fbf.create()
    return fs

# Pipeline executable
#PIPELINE_EXE_SHIS = "/Users/maciek/cvs/devel/Calibrate/Pipeline/L0L1Pipeline_SHISv0"
PIPELINE_EXE_SHIS = "/home/maciek/cvs/devel/Calibrate/Pipeline/L0L1PipelineMmapAdapter_SHISv0"
RADCAL_EXE_SHIS = "/home/maciek/cvs/devel/Calibrate/CalRad/RadCalMmapAdapter_SHISv0"

def new_pipe_ends():
    """ create new pipe, return opened files on both sides """
    rfd, wfd = pipe()
    return fdopen(wfd, 'w'), fdopen(rfd, 'r')

class NewProcessCradle:
    """ Resource wrapper for spawned process that guarantees preservation of environment
    variables and filehandles.

    Initialization spawns a new process, binds pipes to allow for two-way communication 
    via other process's stdin & stdout. """
    
    def __init__(self, exe):
        p2c_write, p2c_read = new_pipe_ends()   # this process -> child
        c2p_write, c2p_read = new_pipe_ends()   # child -> this process
        
        pid = fork()
        if (pid == 0):
            # child process
            p2c_write.close()
            c2p_read.close()

            dup2(p2c_read.fileno(), 0) # replace child process stdin with p2c_out
            dup2(c2p_write.fileno(), 1)  # and stdout with c2p_in
            
            execv(exe, [exe])
            
        else:
            # parent process (This python script)
            p2c_read.close()
            c2p_write.close()
            
            self.sender = p2c_write
            self.recver = c2p_read


class PipelineWrapper:
    """ Object wrapper for running SHISv0 pipeline."""
    
    def __init__(self, band='LW', exe=PIPELINE_EXE_SHIS, data_in="IFG", blackbodyPasses = 2):
        self.band = band
        self.blackbodyPasses = blackbodyPasses # FIXME - this should be self-determined from mirrorPosition file.
        self.bandsize = SHIS_BAND_SIZE[band]

        self.fbfSources = getFbfSources(band, data_in)
        self.numRecords = self.fbfSources['data_in'].fbf.length()

        self.buildMmaps()
        
        # let the child process know about mmaps via the magic of environment variables
        environ['GIPS_BAND'] = self.band
        environ['GIPS_BBPASSES'] = `self.blackbodyPasses`
        environ['GIPS_WN_IN_MMAP'] = `self.fbackers['wn'].fd`
        environ['GIPS_DATA_IN_MMAP'] = `self.fbackers['in'].fd`
        environ['GIPS_TEMPS_IN_MMAP'] = `self.fbackers['temps'].fd`
        environ['GIPS_DATA_OUT_MMAP'] = `self.fbackers['out'].fd`

        self.mmapWriteRec(self.fbfSources['wn_in'].fbf.read(1), 'wn') # make wn available to child.

        #self.sender, self.recvr, self.err_recvr = popen3(exe) # spawn C++ pipeline process
        #self.sender, self.recvr = popen2(exe) # spawn C++ pipeline process
        npc = NewProcessCradle(exe)
        self.sender = npc.sender
        self.recver = npc.recver

    def sendMsg(self, msg):
        #print "Python code sent:" + msg # DEBUG
        print >>self.sender, msg
        self.sender.flush()

    def converse(self, msg):
        self.sendMsg(msg)
        resp=self.recver.readline()
        #print "Python got back:" + resp; # DEBUG
        return resp

    def buildMmaps(self):
        FB_DIR = environ.get('TMP','/tmp')
        
        class MMAP_BACKER:
            def __init__(self, size_in_words, fbf_type):
                self.size = size_in_words
                self.type = fbf_type

        self.fbackers = {'wn': MMAP_BACKER(self.bandsize, 'real8'), 
                         'in': MMAP_BACKER(self.bandsize, 'complex16'), 
                         'out': MMAP_BACKER(self.bandsize, 'real8'), 
                         'temps': MMAP_BACKER(6,'real8')
                         }
        
        for fbufkey in self.fbackers:
            fbuf = self.fbackers[fbufkey]

            # create a memory map for a new FBF file to leverage the FBF library's
            # facility with reading/writing Numeric arrays as binary files.
            fbuf.fbf = FBF(stemname='GIPSTMP_'+fbufkey, typename=fbuf.type, grouping=[fbuf.size,], dirname=FB_DIR)
            fbuf.fp = fbuf.fbf.create()
            fbuf.path = fbuf.fbf.pathname
            fbuf.fd = fbuf.fp.fileno() # convenience method to get file descriptor
            #print "path: %s fd: %s" % (fbuf.path, fbuf.fd) # DEBUG

            # resize buffer to proper size.
            z = zeros(fbuf.size, {'complex16': Complex}.get(fbuf.type, Float))
            fbuf.fbf.block_write(1, z)
            fbuf.fp.flush()
            fbuf.fp.seek(0) 
    
    def mmapWriteRec(self, data, mmapKey, mmapRecord=1):
        """ data is a Numeric or Numarray type object of compatible size/type """
        # use the fact that the mmaps are accessible as FBF objects.
        #print "python writing to mmap %s: %s" % (mmapKey, data) # DEBUG
        self.fbackers[mmapKey].fp.seek(0)
        self.fbackers[mmapKey].fbf.block_write(mmapRecord, data)
        self.fbackers[mmapKey].fp.flush()

    def mmapReadRec(self, mmapKey, mmapRecord=1):
        #print self.fbackers[mmapKey].fbf # DEBUG
        #print self.fbackers[mmapKey].fbf.length() # DEBUG
        return self.fbackers[mmapKey].fbf.read(mmapRecord)

    def createPipeline(self, name):
        reply = self.converse("CREATE PIPELINE:" + name)
        print reply #DEBUG

    def fbfReadRec(self, fileKey, recordNum):
        return self.fbfSources[fileKey].fbf.read(recordNum)

    def fbfReadVal(self, fileKey, recordNum):
        return self.fbfReadRec(fileKey, recordNum)[0]

    def fbfWriteRec(self, data, fileKey, recordNum):
        self.fbfSources[fileKey].fbf.block_write(recordNum, data)

    def cyclePipeline(self, name, recordNum):
        """ run the SHISv0 pipeline for the indicated record """
        self.mmapWriteRec(self.fbfReadRec('data_in', recordNum), 'in')

        temps = zeros(6)    # zero-initialized numarray
        temps[0] = self.fbfReadVal('at_temp', recordNum)
        temps[1] = self.fbfReadVal('ab_temp', recordNum)
        temps[2] = self.fbfReadVal('aa_temp', recordNum)
        temps[3] = self.fbfReadVal('ht_temp', recordNum)
        temps[4] = self.fbfReadVal('hb_temp', recordNum)
        temps[5] = self.fbfReadVal('ha_temp', recordNum)
        self.mmapWriteRec(temps, 'temps')

        time = self.fbfReadVal('time', recordNum)
        obsTypeCode = self.fbfReadVal('obsType', recordNum)
        SHIS_OBS_TYPE_MAP = {
            63: 'INVALID',      # invalid entry
            65: 'AMBIENT',    # ambient blackbody
            69: 'OBS',    # earth observation
            72: 'HOT',    # hot blackbody
            83: 'SPACE'     # sky observation
            }

        obsType = SHIS_OBS_TYPE_MAP.get(obsTypeCode, None)
        if not obsType:
            print "Record %s has invalid obs type" % recordNum
            return 

        # run the C++ pipeline: Tas
        response = self.converse("PROCESS PIPELINE:" + name +
                                 " TASK:" + `recordNum` +
                                 " time:" + `time` +
                                 " obsType:" + obsType)
        
        # process the response
        respKeyVals = response.split() 
        outRecord = None
        # response consists of key:value pairs separated by spaces.
        for rkv in respKeyVals:
            k, v = rkv.split(':')
            if k == 'TASK':
                outRecord = int(v)
            # FIXME - insert here more audit code to read output from C++ pipeline

        if not outRecord:
            pass  # waiting - no "TASK" keyword appeared

        else: 
            outData = self.mmapReadRec('out')
            if GPLOT_ON:
                gplot.plot(outData) ### GPLOT ### - remove if not installed.
            self.fbfWriteRec(outData, 'data_out', outRecord)

def runPipeline(beginRec=1, numRecords=None, executable=PIPELINE_EXE_SHIS, fbf_in="IFG", blackbodyPasses = 2):
    pw = PipelineWrapper(exe=executable, data_in=fbf_in, blackbodyPasses = blackbodyPasses)
    pw.createPipeline("FW")
    pw.createPipeline("BW")

    if not numRecords or beginRec + numRecords > pw.numRecords:
        numRecords = pw.numRecords - beginRec + 1
    
    for rec in range(beginRec, beginRec + numRecords):
        scanDirection = {66:'BW', 70:'FW'}.get(pw.fbfReadVal('scanDir', rec), None) # default FW
        if not scanDirection:
            print "Invalid scan direction for record %s" % rec
            # FIXME - insert logic to fill record with empty... somehow tag it as invalid
            continue

        pw.cyclePipeline(scanDirection, rec)
    
    # cleanup
    for fbfObj in pw.fbfSources.values():
        fbfObj.fbf.close()

def usage():
    return """Usage: runPipeline [-hxrn] [firstRecord=1] [numRecords=max]
    -h: this help message
    -x execname: use execname as pipeline to run instead of default.
    -r: run only radiometric calibration, using raw SpectrumXW.complex8.yyyy input
    -n: run only radiometric calibration using nonlinearity-corrected SpectrumXW_NLC.complex8.yyy input
    -b N: each blackbody stared at N times per sweep (default: 2 - such as in Proteus flights)
    """
    
if __name__ == '__main__':
    import sys
    from getopt import getopt, GetoptError

    ### GPLOT ### - remove if not installed.
    if GPLOT_ON:
        import Gnuplot
        gplot = Gnuplot.Gnuplot()
        gplot.plot([1,2,3,4,5])
        gplot.set_range('yrange',[0,200])
        gplot.set_range('xrange',[0,2044])
        print "Gnuplot activated"
    ### GPLOT ###

    try:
        opts, args = getopt(sys.argv[1:], 'hx:rnb:')
    except GetoptError:
        print usage()
        sys.exit(0)

    executable = PIPELINE_EXE_SHIS
    fbf_in = "IFG"
    blackbodyPasses = 2

    for o, a in opts:
        if o == '-h':
            print usage()
            sys.exit(0)
        if o == '-x':
            executable = a
        if o == '-r':
            executable = RADCAL_EXE_SHIS
            fbf_in = "SPEC_RAW"
        if o == '-n':
            executable = RADCAL_EXE_SHIS
            fbf_in = "SPEC_NLC"
        if o == '-b':
            blackbodyPasses = int(a)
            
    if len(args) == 0:
        runPipeline(executable=executable, fbf_in=fbf_in, blackbodyPasses = blackbodyPasses)
    elif len(args) == 1:
        runPipeline(beginRec=int(args[0]),
                    executable=executable, fbf_in=fbf_in, blackbodyPasses = blackbodyPasses)
    elif len(args) == 2:
        runPipeline(beginRec=int(args[0]), numRecords=int(args[1]),
                    executable=executable, fbf_in=fbf_in, blackbodyPasses = blackbodyPasses)
    else: 
        print usage()
        sys.exit(0)
